// Copyright (C) 2019-2021  Nexedi SA and Contributors.
//                          Kirill Smelkov <kirr@nexedi.com>
//
// This program is free software: you can Use, Study, Modify and Redistribute
// it under the terms of the GNU General Public License version 3, or (at your
// option) any later version, as published by the Free Software Foundation.
//
// You can also Link and Combine this program with other software covered by
// the terms of any of the Free Software licenses or any of the Open Source
// Initiative approved licenses and Convey the resulting work. Corresponding
// source of such a combination shall include the source code for all other
// software used.
//
// This program is distributed WITHOUT ANY WARRANTY; without even the implied
// warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
//
// See COPYING file for full licensing terms.
// See https://www.nexedi.com/licensing for rationale and options.

package zdata
// tests for δftail.go
//
// This are the main tests for ΔFtail functionality. The primary testing
// concern is to verify how ΔFtail merges ΔBtail and ΔZtail histories on Update
// and queries.
//
// We assume that ΔBtail works correctly (this is covered by ΔBtail tests)
// -> no need to exercise many different topologies and tracking sets.
//
// Since ΔFtail does not recompute anything by itself when tracking set
// changes, and only merges δBtail and δZtail histories on queries, there is no
// need to exercise many different tracking sets. Once again we assume that
// ΔBtail works correctly and verify δFtail only with track=[-∞,∞).
//
// There are 2 testing approaches:
//
//   a) transition a ZBigFile in ZODB through particular .blktab and ZBlk
//      states and feed ΔFtail through created database transactions.
//   b) transition a ZBigFile in ZODB through   random   .blktab and ZBlk
//      states and feed ΔFtail through created database transactions.
//
// TestΔFtail and TestΔFtailRandom implement approaches "a" and "b" correspondingly.

import (
	"context"
	"fmt"
	"reflect"
	"sort"
	"strings"
	"testing"

	"lab.nexedi.com/kirr/go123/exc"
	"lab.nexedi.com/kirr/neo/go/transaction"
	"lab.nexedi.com/kirr/neo/go/zodb"

	"lab.nexedi.com/nexedi/wendelin.core/wcfs/internal/set"
	"lab.nexedi.com/nexedi/wendelin.core/wcfs/internal/xbtree/xbtreetest"
)

type setStr = set.Str

const ø = "ø"

// ΔFTestEntry represents one entry in ΔFtail tests.
type ΔFTestEntry struct {
	δblkTab  map[int64]string  // changes in tree part  {} #blk -> ZBlk<name>
	δdataTab setStr            // changes to ZBlk objects
}

// TestΔFtail runs ΔFtail tests on set of concrete prepared testcases.
func TestΔFtail(t *testing.T) {
	// δT is shorthand to create δblkTab.
	type δT = map[int64]string

	// δD is shorthand to create δdataTab.
	δD := func(zblkv ...string) setStr {
		δ := setStr{}
		for _, zblk := range zblkv {
			δ.Add(zblk)
		}
		return δ
	}

	const a,b,c,d,e,f,g,h,i,j = "a","b","c","d","e","f","g","h","i","j"

	testv := []ΔFTestEntry{
		{δT{1:a,2:b,3:ø},	δD(a)},
		{δT{},			δD(c)},
		{δT{2:c},		δD(a,b)},

		// clear the tree
		{δT{1:ø,2:ø},		δD()},

		// i is first associated with file, but later unlinked from it
		// then i is changed -> the file should no be in δF
		{δT{5:i},		δD()},
		{δT{5:e},		δD()},
		{δT{},			δD(i)},

		// XXX text
		{nil, nil},

		// ---- found by TestΔFtailRandom ----

		{δT{1:a,6:i,7:d,8:e},   δD(a,c,e,f,g,h,i,j)},

		// was including <= lo entries in SliceByFileRev
		{δT{0:b,2:j,3:i,5:f,6:b,7:i,8:d},	δD(a,b,c,d,e,g,i,j)},
		{δT{0:e,2:h,4:d,9:b},			δD(a,h,i)},
		{δT{0:j,1:i,3:g,5:a,6:e,7:j,8:f,9:d},	δD()},
		{δT{0:b,1:f,2:h,4:b,8:b},		δD(b,d,i)},
		{δT{1:a,3:d,6:j},			δD(b,c,d,f,g,h,i,j)},
		{δT{0:i,1:f,4:e,5:e,7:d,8:h},		δD(d,j)},
		{δT{},					δD(a,b,c,e,f,g,h,i,j)},
	}

	testq := make(chan ΔFTestEntry)
	go func() {
		defer close(testq)
		for _, test := range testv {
			testq <- test
		}
	}()
	testΔFtail(t, testq)
}

// TestΔFtailRandom runs ΔFtail tests on randomly-generated file changes.
func TestΔFtailRandom(t *testing.T) {
	n    := xbtreetest.N(1E3, 1E4, 1E5)
	nblk := xbtreetest.N(1E1, 2E1, 1E2) // keeps failures detail small on -short

	// random-number generator
	rng, seed := xbtreetest.NewRand()
	t.Logf("# n=%d seed=%d", n, seed)

	vv := "abcdefghij"
	randv := func() string {
		i := rng.Intn(len(vv))
		return vv[i:i+1]
	}

	testq := make(chan ΔFTestEntry)
	go func() {
		defer close(testq)
		for i := 0; i < n; i++ {
			nδblkTab  := rng.Intn(nblk)
			nδdataTab := rng.Intn(len(vv))

			δblkTab  := map[int64]string{}
			δdataTab := setStr{}

			blkv := rng.Perm(nblk)
			for j := 0; j < nδblkTab; j++ {
				blk  := blkv[j]
				zblk := randv()
				δblkTab[int64(blk)] = zblk
			}

			vv_ := rng.Perm(len(vv))
			for j := 0; j < nδdataTab; j++ {
				k := vv_[j]
				v := vv[k:k+1]
				δdataTab.Add(v)
			}

			testq <- ΔFTestEntry{δblkTab, δdataTab}
		}
	}()

	testΔFtail(t, testq)
}

// testΔFtail verifies ΔFtail on sequence on testcases coming from testq.
func testΔFtail(t_ *testing.T, testq chan ΔFTestEntry) {
	t := xbtreetest.NewT(t_)
	X := exc.Raiseif

	xat := map[zodb.Tid]string{} // tid > "at<i>"

	// start δFtail when zfile does not yet exists
	// this way we'll verify how ΔFtail rebuilds vδE for started-to-be-tracked file
	t0 := t.CommitTree("øf")
	xat[t0.At] = "at0"
	t.Logf("#   @at0 (%s)", t0.At)
	δFtail := NewΔFtail(t.Head().At, t.DB)

	// load dataTab
	dataTab  := map[string]string{}               // ZBlk<name> -> data
	for /*oid*/_, zblki := range t.Head().ZBlkTab {
		dataTab[zblki.Name] = zblki.Data
	}

	// create zfile, but do not track it yet
	t1 := t.CommitTree(fmt.Sprintf("t0:a D%s", dataTabTxt(dataTab)))
	xat[t1.At] = "at1"
	t.Logf("# → @at1 (%s)  %s\t; not-yet-tracked", t1.At, t1.Tree)
	δF, err := δFtail.Update(t1.ΔZ);  X(err)
	if !(δF.Rev == t1.At && len(δF.ByFile) == 0) {
		t.Errorf("wrong δF:\nhave {%s, %v}\nwant: {%s, ø}", δF.Rev, δF.ByFile, t1.At)
	}

	// load zfile via root['treegen/file']
	txn, ctx := transaction.New(context.Background())
	zconn, err := t.DB.Open(ctx, &zodb.ConnOptions{At: t.Head().At, NoPool: true});  X(err)
	xzroot, err := zconn.Get(ctx, 0);  X(err)
	zroot := xzroot.(*zodb.Map)
	err = zroot.PActivate(ctx);  X(err)
	zfile := zroot.Data["treegen/file"].(*ZBigFile)
	zroot.PDeactivate()
	foid := zfile.POid()
	err = zfile.PActivate(ctx);  X(err)
	blksize   := zfile.blksize
	blktabOid := zfile.blktab.POid()
	if blktabOid != t.Root() {
		t.Fatalf("BUG: zfile.blktab (%s)  !=  treeroot (%s)", blktabOid, t.Root())
	}
	zfile.PDeactivate()

	// start track zfile[0,∞) from the beginning
	// this should make ΔFtail to see all zfile changes
	size, path, err := zfile.Size(ctx);  X(err)
	δFtail.Track(zfile, /*blk*/-1, path, /*zblk*/nil)
	if sizeOK := 1*blksize; size != sizeOK { // NOTE maches t1 commit
		t.Fatalf("BUG: zfile size: have %d  ; want %d", size, sizeOK)
	}

	// data built via applying changes from testv
	vδf      := []*ΔFile{                         // (rev↑, {}blk)
			{Rev: t1.At, Epoch: true},
		    }
	vδE      := []_ΔFileEpoch{                    // (rev↑, EPOCH)
			{
				Rev:             t1.At,
				oldRoot:         zodb.InvalidOid,
				newRoot:         blktabOid,
				newBlkSize:      blksize,
				oldTrackSetZBlk: nil,
			},
		    }
	blkTab   := map[int64]string{0:"a"}           // #blk -> ZBlk<name>
	Zinblk   := map[string]setI64{}               // ZBlk<name> -> which #blk refer to it
	blkRevAt := map[zodb.Tid]map[int64]zodb.Tid{} // {} at -> {} #blk -> rev

	// retrack should be called after new epoch to track zfile[-∞,∞) again
	retrack := func() {
		for blk := range blkTab {
			_, path, zblk, _, err := zfile.LoadBlk(ctx, blk);  X(err)
			δFtail.Track(zfile, blk, path, zblk)
		}
	}


	epochv := []zodb.Tid{t0.At, t1.At}

	// δfstr/vδfstr converts δf/vδf to string taking xat into account
	δfstr := func(δf *ΔFile) string {
		s := fmt.Sprintf("@%s·%s", xat[δf.Rev], δf.Blocks)
		if δf.Epoch {
			s += "E"
		}
		if δf.Size {
			s += "S"
		}
		return s
	}
	vδfstr := func(vδf []*ΔFile) string {
		var s []string
		for _, δf := range vδf {
			s = append(s, δfstr(δf))
		}
		return fmt.Sprintf("%s", s)
	}


	i := 1 // matches t1
	delfilePrev := false
	for test := range testq {
		i++
		δblk := setI64{}
		δtree := false
		delfile := false

		// command to delete zfile
		if test.δblkTab == nil && test.δdataTab == nil {
			delfile = true
		}

		// new epoch starts when file is deleted or recreated
		newEpoch := delfile || (!delfile && delfile != delfilePrev)
		delfilePrev = delfile

		ZinblkPrev := map[string]setI64{}
		for zblk, inblk := range Zinblk {
			ZinblkPrev[zblk] = inblk.Clone()
		}

		// newEpoch -> reset
		if newEpoch {
			blkTab = map[int64]string{}
			Zinblk = map[string]setI64{}
			δblk   = nil
		} else {
			// rebuild blkTab/Zinblk
			for blk, zblk := range test.δblkTab {
				zprev, ok := blkTab[blk]
				if ok {
					delete(Zinblk[zprev], blk)
				} else {
					zprev = ø
				}

				if zblk != ø {
					blkTab[blk] = zblk
					inblk, ok := Zinblk[zblk]
					if !ok {
						inblk = setI64{}
						Zinblk[zblk] = inblk
					}
					inblk.Add(blk)
				} else {
					delete(blkTab, blk)
				}

				// update δblk due to change in blkTab
				if zblk != zprev {
					δblk.Add(blk)
					δtree = true
				}
			}

			// rebuild dataTab
			for zblk := range test.δdataTab {
				data, ok := dataTab[zblk]               // e.g. a -> a2
				if !ok {
					t.Fatalf("BUG: blk %s not in dataTab\ndataTab: %v", zblk, dataTab)
				}
				data = fmt.Sprintf("%s%d", data[:1], i) // e.g. a4
				dataTab[zblk] = data

				// update δblk due to change in ZBlk data
				for blk := range Zinblk[zblk] {
					δblk.Add(blk)
				}
			}
		}

		// commit updated zfile / blkTab + dataTab
		var req string
		if delfile {
			req = "øf"
		} else {
			tTxt := "t" + xbtreetest.KVTxt(blkTab)
			dTxt := "D" + dataTabTxt(dataTab)
			req  = tTxt + " " + dTxt
		}
		commit := t.CommitTree(req)
		if newEpoch {
			epochv = append(epochv, commit.At)
		}
		xat[commit.At] = fmt.Sprintf("at%d", i)
		flags := ""
		if newEpoch {
			flags += "\tEPOCH"
		}
		t.Logf("# → @%s (%s)  δT%s  δD%s\t; %s\tδ%s%s", xat[commit.At], commit.At, xbtreetest.KVTxt(test.δblkTab), test.δdataTab, commit.Tree, δblk, flags)
		//t.Logf("#   vδf: %s", vδfstr(vδf))


		// update blkRevAt
		var blkRevPrev map[int64]zodb.Tid
		if i != 0 {
			blkRevPrev = blkRevAt[δFtail.Head()]
		}
		blkRev := map[int64]zodb.Tid{}
		for blk, rev := range blkRevPrev {
			if newEpoch {
				blkRev[blk] = commit.At
			} else {
				blkRev[blk] = rev
			}
		}
		for blk := range δblk {
			blkRev[blk] = commit.At
		}
		blkRevAt[commit.At] = blkRev
		if false {
			fmt.Printf("blkRevAt[@%s]:\n", xat[commit.At])
			blkv := []int64{}
			for blk := range blkRev {
				blkv = append(blkv, blk)
			}
			sort.Slice(blkv, func(i, j int) bool {
				return blkv[i] < blkv[j]
			})
			for _, blk := range blkv {
				fmt.Printf("  #%d:  %v\n", blk, blkRev[blk])
			}
		}

		// update zfile
		txn.Abort()
		txn, ctx = transaction.New(context.Background())
		err = zconn.Resync(ctx, commit.At);  X(err)

		var δfok *ΔFile
		if newEpoch || len(δblk) != 0 {
			δfok = &ΔFile{
				Rev:    commit.At,
				Epoch:  newEpoch,
				Blocks: δblk,
				Size:   δtree, // not strictly ok, but matches current ΔFtail code
			}
			vδf = append(vδf, δfok)
		}
		if newEpoch {
			δE := _ΔFileEpoch{Rev: commit.At}
			if delfile {
				δE.oldRoot    = blktabOid
				δE.newRoot    = zodb.InvalidOid
				δE.newBlkSize = -1
				// XXX oldBlkSize ?
			} else {
				δE.oldRoot    = zodb.InvalidOid
				δE.newRoot    = blktabOid
				δE.newBlkSize = blksize
				// XXX oldBlkSize ?
			}
			oldTrackSetZBlk := map[zodb.Oid]setI64{}
			for zblk, inblk := range ZinblkPrev {
				oid, _ := commit.XGetBlkByName(zblk)
				oldTrackSetZBlk[oid] = inblk
			}
			δE.oldTrackSetZBlk = oldTrackSetZBlk
			vδE = append(vδE, δE)
		}

		//fmt.Printf("Zinblk: %v\n", Zinblk)

		// update δFtail
		δF, err := δFtail.Update(commit.ΔZ);  X(err)

		// assert δF points to the file if δfok != ø
		if δF.Rev != commit.At {
			t.Errorf("wrong δF.Rev: have %s  ; want %s", δF.Rev, commit.At)
		}
		δfiles := setOid{}
		for δfile := range δF.ByFile {
			δfiles.Add(δfile)
		}
		δfilesOK := setOid{}
		if δfok != nil {
			δfilesOK.Add(foid)
		}
		if !δfiles.Equal(δfilesOK) {
			t.Errorf("wrong δF.ByFile:\nhave keys: %s\nwant keys: %s", δfiles, δfilesOK)
			continue
		}

		// verify δf
		δf := δF.ByFile[foid]
		if !reflect.DeepEqual(δf, δfok) {
			t.Errorf("δf:\nhave: %v\nwant: %v", δf, δfok)
		}

		// track whole zfile again if new epoch was started
		if newEpoch {
			retrack()
		}

		// verify δFtail.trackSetZBlk
		trackZinblk := map[string]setI64{}
		for oid, zt := range δFtail.trackSetZBlk {
			zblki := commit.ZBlkTab[oid]
			for root, blocks := range zt.inroot {
				if root != blktabOid {
					t.Errorf(".trackSetZBlk: zblk %s points to unexpected blktab %s", zblki.Name, blktabOid)
					continue
				}

				inblk, ok := trackZinblk[zblki.Name]
				if !ok {
					inblk = setI64{}
					trackZinblk[zblki.Name] = inblk
				}
				inblk.Update(blocks)
			}
		}
		if !reflect.DeepEqual(trackZinblk, Zinblk) {
			t.Errorf(".trackSetZBlk:\n~have: %v\n want: %v", trackZinblk, Zinblk)
		}

		// ForgetPast configured threshold
		const ncut = 5
		if len(vδf) >= ncut {
			revcut := vδf[0].Rev
			t.Logf("# forget ≤ @%s", xat[revcut])
			δFtail.ForgetPast(revcut)
			vδf = vδf[1:]
			//t.Logf("#   vδf: %s", vδfstr(vδf))
			//t.Logf("#   vδt: %s", vδfstr(δFtail.SliceByFileRev(zfile, δFtail.Tail(), δFtail.Head())))
			icut := 0;
			for ; icut < len(vδE); icut++ {
				if vδE[icut].Rev > revcut {
					break
				}
			}
			vδE = vδE[icut:]
		}

		// verify δftail.root
		δftail := δFtail.byFile[foid]
		rootOK := blktabOid
		if delfile {
			rootOK = zodb.InvalidOid
		}
		if δftail.root != rootOK {
			t.Errorf(".root: have %s  ; want %s", δftail.root, rootOK)
		}

		// verify vδE
		if !reflect.DeepEqual(δftail.vδE, vδE) {
			t.Errorf("vδE:\nhave: %v\nwant: %v", δftail.vδE, vδE)
		}


		// SliceByFileRev
		for j := 0; j < len(vδf); j++ {
			for k := j; k < len(vδf); k++ {
				var lo zodb.Tid
				if j == 0 {
					lo = vδf[0].Rev - 1
				} else {
					lo = vδf[j-1].Rev
				}
				hi := vδf[k].Rev

				vδf_ok := vδf[j:k+1] // [j,k]
				vδf_ := δFtail.SliceByFileRev(zfile, lo, hi)
				if !reflect.DeepEqual(vδf_, vδf_ok) {
					t.Errorf("slice (@%s,@%s]:\nhave: %v\nwant: %v", xat[lo], xat[hi], vδfstr(vδf_), vδfstr(vδf_ok))
				}
			}
		}


		// BlkRevAt

		blkv := []int64{} // all blocks
		if l := len(vδf); l > 0 {
			for blk := range blkRevAt[vδf[l-1].Rev] {
				blkv = append(blkv, blk)
			}
		}
		blkv = append(blkv, 1E4/*this block is always hole*/)
		sort.Slice(blkv, func(i, j int) bool {
			return blkv[i] < blkv[j]
		})

		for j := 0; j < len(vδf); j++ {
			at := vδf[j].Rev
			blkRev := blkRevAt[at]
			for _, blk := range blkv {
				rev, exact := δFtail.BlkRevAt(ctx, zfile, blk, at)
				revOK, ok := blkRev[blk]
				if !ok {
					k := len(epochv) - 1
					for ; k >= 0; k-- {
						if epochv[k] <= at {
							break
						}
					}
					revOK = epochv[k]
				}
				exactOK := true
				if revOK <= δFtail.Tail() {
					revOK, exactOK = δFtail.Tail(), false
				}
				if !(rev == revOK && exact == exactOK) {
					t.Errorf("blkrev #%d @%s:\nhave: @%s, %v\nwant: @%s, %v", blk, xat[at], xat[rev], exact, xat[revOK], exactOK)
				}
			}

		}
	}
}


// dataTabTxt returns string representation of {} dataTab.
func dataTabTxt(dataTab map[string]string) string {
	// XXX dup wrt xbtreetest.KVTxt but uses string instead of Key for keys.
	if len(dataTab) == 0 {
		return "ø"
	}

	keyv := []string{}
	for k := range dataTab { keyv = append(keyv, k) }
	sort.Strings(keyv)

	sv := []string{}
	for _, k := range keyv {
		v := dataTab[k]
		if strings.ContainsAny(v, " \n\t,:") {
			panicf("[%v]=%q: invalid value", k, v)
		}
		sv = append(sv, fmt.Sprintf("%v:%s", k, v))
	}

	return strings.Join(sv, ",")
}
